from __future__ import annotations
import sys

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

_ELEMENT_MASS = {
    "H": 1.00794,
    "Be": 9.0122,
    "C": 12.0107,
    "N": 14.0067,
    "O": 15.9994,
    "Ca": 40.078,
    "S": 32.065,
}

_VAR_PREFIXES = [
    # These are checked in the order written, so should be ordered by most specific to least specific (eg. concN before conc).
    "vmr",
    "mmr",
    "sconc",
    "concNt",
    "concN",
    "concC",
    "conc",
    "wet",
    "dry",
    "proxydry",
    "proxywet",
    "dep",
]

_VAR_POSTFIXES = ["pr"]

_MOLMASSES = {
    # Override molecular masses used for species that do not constitute molecular
    # formulas with only single letter elements.
    "air_dry": 28.9647,
    "isop": 68.12,
    "glyoxal": 58.036,
    "glyox": 58.036,
}


class UnknownSpeciesError(ValueError):
    pass


def _get_species(aerocom_var: str) -> str:
    """
    Get species name from variable name

    Parameters
    ----------
    var_name : str
        pyaerocom variable name (cf. variables.ini)

    Raises
    ------
    UnknownSpeciesError
        if species cannot be inferred

    Returns
    -------
    str
        name of species

    """
    if aerocom_var in _MOLMASSES:
        return aerocom_var
    species = None
    for prefix in _VAR_PREFIXES:
        if aerocom_var.startswith(prefix):
            species = aerocom_var.split(prefix)[-1]
            break
            # if species in _MOLMASSES:

    if species is None:
        raise UnknownSpeciesError(
            f"Could not infer atom / molecule/ species from var_name {aerocom_var}"
        )
    for postfix in _VAR_POSTFIXES:
        if species.endswith(postfix):
            species = species.split(postfix)[0]
            break

    return species


class MolecularMass:
    """This class represents the mass of simple molecular formulas
    and permits arithmetic calculations on them.
    """

    def __init__(self, val: str | int | float, *, label: str | None = None):
        """
        :param val: Either a chemical formula (eg. H2O) or a numeric mass.
        :param label: Optional label. Defaults to the chemical formula / chemical mass.

        :raises ValueError: If val is negative.
        """
        if isinstance(val, float | int):
            if val < 0:
                raise ValueError("Molecular mass can not be less than zero.")
            self._mass = float(val)
        else:
            self._mass = self._mass_from_chemical_formula(val)

        if isinstance(val, str) and label is None:
            self._label: str | None = val
        else:
            self._label = label

    @property
    def mass(self) -> float:
        """
        :return: The mass value (in Dalton).
        """
        return self._mass

    @property
    def label(self) -> str:
        """
        :return: Human readable description of the mass (eg. '1 u' or 'H2O')
        """
        if self._label is None:
            return f"{self.mass:.4f} u"

        return self._label

    def _mass_from_chemical_formula(self, val: str) -> float:
        """Determines the mass of simple chemical formulas.

        :param val: Chemicala formula as string (see limitations below).
        :raises ValueError: If there are brackets in the formula.
        :raises ValueError: If element masses are not defined in _ELEMENT_MASSES dict.

        :return: Molecular mass (in daltons).

        Limitations:
        ------------
        - Elements must have their first character and only the first capitalized.
        - Numeric subscripts are limited to one digit.
        - No brackets.
        """
        # Limitations
        # Does not support brackets (eg. Ca(OH)2).
        # Does not support multidigit numbers (eg. C6H12O6).
        if "(" in val or ")" in val:
            raise ValueError("Brackets are not currently supported.")
        mass: float = 0
        i = 0
        while i < len(val):
            offset = 0
            element = val[i]
            if i + 1 < len(val):
                if val[i + 1].islower():
                    element += val[i + 1]
                    offset += 1
            if i + offset + 1 < len(val):
                if val[i + offset + 1].isdigit():
                    count = int(val[i + offset + 1])
                    offset += 1
                else:
                    count = 1
            else:
                count = 1

            try:
                mass += count * _ELEMENT_MASS[element]
            except KeyError as e:
                raise ValueError(f"Unable to parse chemical formula '{val}'") from e

            i += offset + 1

        return mass

    def __add__(self, other) -> MolecularMass:
        if isinstance(other, MolecularMass):
            return MolecularMass(self.mass + other.mass)

        return MolecularMass(self.mass + other)

    def __radd__(self, other) -> MolecularMass:
        return self.__add__(other)

    def __sub__(self, other) -> MolecularMass:
        if isinstance(other, MolecularMass):
            return MolecularMass(self.mass - other.mass)

        return MolecularMass(self.mass - other)

    def __mul__(self, other) -> float:
        if isinstance(other, MolecularMass):
            self.mass * other.mass

        return self.mass * other

    def __rmul__(self, other) -> float:
        return self.__mul__(other)

    def __truediv__(self, other) -> float:
        if isinstance(other, MolecularMass):
            return self.mass / other.mass

        return self.mass / other

    def __float__(self) -> float:
        return float(self.mass)

    @override
    def __repr__(self) -> str:
        return f"MolecularMass('{self.label}')"

    @override
    def __str__(self) -> str:
        return self.label

    @staticmethod
    def from_aerocom_var(aerocom_var: str) -> MolecularMass:
        """Produces a MolecularMass from an var_name. It will first check
        the override dict _MOLMASSES. If the var_name is not defined there
        it will try to extract the chemical formula from the name (eg. 'concso2'->'SO2').

        :param aerocom_var: string (See variables.ini).

        :return: MolecularMass of element.
        """
        species = _get_species(aerocom_var)
        if species in _MOLMASSES:
            return MolecularMass(_MOLMASSES[species], label=species)
        return MolecularMass(species.upper())


def get_molmass(aerocom_varname: str) -> float:
    """
    Get molar mass for input variable

    Parameters
    ----------
    var_name : str
        pyaerocom variable name (cf. variables.ini) or name of species

    Returns
    -------
    float
        molar mass of species in units of g/mol

    Note:
    Will break with molecular formulas that contain 2-letter elements (eg. Ca)
    since this relies on upper case conversion to make it understandable for
    MolecularMass.__init__(). For now such species should be defined manually
    in the override dict _MOLMASSES.
    """
    try:
        mass = MolecularMass.from_aerocom_var(aerocom_varname).mass
    except UnknownSpeciesError:
        # This function previously was used on pure formulas even though it is only
        # supposed to be used on var_names.
        # This maintains backward compatibility for those cases.
        mass = MolecularMass(aerocom_varname.upper()).mass

    return mass


def get_mmr_to_vmr_fac(aerocom_varname: str) -> float:
    """
    Get conversion factor for MMR -> VMR conversion for input variable

    Note
    ----
    Assumes dry air molar mass

    Parameters
    ----------
    var_name : str
        Name of variable to be converted

    Returns
    -------
    float
        multiplication factor to convert MMR -> VMR
    """
    return get_molmass("air_dry") / get_molmass(aerocom_varname)



from __future__ import annotations

from datetime import date, datetime, timedelta
from pathlib import Path

import pytest
import xarray as xr

from pyaerocom.io.cams2_83.models import ModelName, RunType
from pyaerocom.io.cams2_83.reader import AEROCOM_NAMES, DATA_FOLDER_PATH
from pyaerocom.io.cams2_83.reader import model_paths as find_model_paths
from pyaerocom.io.cams2_83.reader import read_dataset

TEST_DATE = datetime(2021, 12, 1)
TEST_DATES = [TEST_DATE + timedelta(days=d) for d in range(3)]


@pytest.fixture
def model_paths(
    model: str | ModelName, run: str | RunType, dates: list[str | date | datetime]
) -> list[Path]:
    if not DATA_FOLDER_PATH.is_dir():
        pytest.skip(f"no access to {DATA_FOLDER_PATH}")
    paths = find_model_paths(model, *dates, run=run)
    return list(paths)


@pytest.mark.parametrize("model", ["EMEP"])
@pytest.mark.parametrize("run", RunType)
@pytest.mark.parametrize(
    "dates,num,unique",
    [
        ([TEST_DATE, TEST_DATE.date(), TEST_DATE.strftime("%Y%m%d")], 3, 1),
        (TEST_DATES, len(TEST_DATES), len(TEST_DATES)),
    ],
)
def test_model_paths(model_paths: list[Path], num: int, unique: int):
    assert bool(model_paths)
    assert all(path.exists() for path in model_paths)
    assert len(model_paths) == num
    assert len(set(model_paths)) == unique


@pytest.fixture
def model_dataset(model_paths: list[Path], day: int) -> xr.Dataset:
    return read_dataset(model_paths, day=day)


@pytest.mark.parametrize("dates,steps", [(TEST_DATES, len(TEST_DATES) * 24)])
@pytest.mark.parametrize("model", ["EMEP"])
@pytest.mark.parametrize("run,day", [("AN", 0), ("FC", 0), ("FC", 3)])
def test_model_file_contents(model_dataset: xr.Dataset, steps: int):
    for var_name in AEROCOM_NAMES.values():
        assert var_name in model_dataset
    assert len(model_dataset.time) == steps



from __future__ import annotations

from pathlib import Path

import pytest

from pyaerocom import const
from pyaerocom.io.icos.reader import ReadICOS

if not const.has_access_lustre:
    pytestmark = pytest.skip(
        reason="Skipping tests that require access to AEROCOM database on METNo servers",
        allow_module_level=True,
    )

VARS_DEFAULT = {"vmrco2", "vmrch4", "vmrco"}
VARS_PROVIDED = VARS_DEFAULT  # | {} add more if ever needed

station_names = pytest.mark.parametrize("station", ("Birkenes", "Gartow", "Hohenpeissenberg"))


@pytest.fixture(scope="module")
def icos_path() -> Path:
    try:
        return Path(const.OBSLOCS_UNGRIDDED[const.ICOS_NAME])
    except KeyError:
        pytest.skip(reason="ICOS path not initialised due to non existence in CI")


@pytest.fixture(scope="module")
def reader(icos_path: Path) -> ReadICOS:
    return ReadICOS(data_dir=str(icos_path))


@pytest.fixture()
def station_files(icos_path: Path, station: str) -> list[Path]:
    files = sorted(icos_path.rglob(f"icos-co2-{station}*.nc"))
    assert files, f"no files for {station}"
    return files


def test_DATASET_NAME(reader: ReadICOS):
    assert reader.DATASET_NAME == "ICOS"


def test_DEFAULT_VARS(reader: ReadICOS):
    assert set(reader.DEFAULT_VARS) >= VARS_DEFAULT


def test_files(reader: ReadICOS):
    assert reader.files, "no stations files found"
    assert len(reader.files) >= 1, "found less files than expected"


def test_FOUND_FILES(reader: ReadICOS):
    assert reader.FOUND_FILES, "no stations files found"
    assert len(reader.FOUND_FILES) >= 1, "found less files than expected"


@station_names
def test_stations(reader: ReadICOS, station: str):
    assert reader.stations()[station], f"no {station} station files"


def test_PROVIDES_VARIABLES(reader: ReadICOS):
    assert set(reader.PROVIDES_VARIABLES) >= VARS_PROVIDED


@station_names
def test_read_file(reader: ReadICOS, station_files: list[str]):
    data = reader.read_file(station_files[-1])
    assert set(data.contains_vars) <= VARS_DEFAULT


def test_read_file_error(reader: ReadICOS):
    bad_station_file = "not-a-file.nc"
    with pytest.raises(ValueError) as e:
        reader.read_file(bad_station_file)
    assert str(e.value) == f"missing {bad_station_file}"


@station_names
def test_read(reader: ReadICOS, station_files: list[str]):
    for var in VARS_PROVIDED:
        data = reader.read(var, station_files, first_file=0, last_file=1)
        assert set(data.contains_vars) <= VARS_PROVIDED


def test_read_error(reader: ReadICOS):
    bad_variable_name = "not-a-variable"
    with pytest.raises(ValueError) as e:
        reader.read(bad_variable_name)
    assert str(e.value) == f"Unsupported variables: {bad_variable_name}"


def test_reader_gives_correct_icos_path(reader: ReadICOS, icos_path: Path):
    assert reader.data_dir == str(icos_path)



import numpy as np
from numpy.typing import DTypeLike


class DynamicRecArrayException(Exception):
    pass


class DynamicRecArray:
    def __init__(self, dtype: DTypeLike, capacity: int = 10):
        """A dynamic record based array of type dtype.

        :param dtype: Datatype of the array. Must be readable by :class:`~numpy.dtype`
        :param capacity: optional initial capacity, i.e. initial hidden array-size.
            Use a large capacity if you intent use many `append` operation on the data.
        """
        self._dtype = np.dtype(dtype)
        self._length = 0
        self._capacity = capacity
        self._array = np.empty(self._capacity, dtype=self._dtype)

    def __len__(self):
        return self._length

    def keys(self):
        """all available data-fields, excluding variable and units which are
        considered metadata"""
        return self._array.dtype.names

    def append(self, rec: list[tuple] | np.ndarray):
        """append this dataset with a record of the same type

        Example: append two rows to a three-datatype array
        ary.append([(0, 1, 3),
                    (4, 5, 6)])

        :param rec: a numpy array of the same datatype, or a list of tuples with the same number of elements

        """
        newlength = self._length + len(rec)
        resize = False
        while self._capacity <= newlength:
            self._capacity += 10 + (self._capacity >> 3)  # 20 + 1.125*self._capacity
            resize = True
        if resize:
            self._array = np.resize(self._array, self._capacity)
        self._array[self._length : newlength] = rec
        self._length = newlength

    def append_array(self, **kwargs):
        """append data using a dictionary of np-arrays

        :raises DynamicRecArrayException: if keys are missing, or shapes are not equal
        """
        key0 = self.keys()[0]
        for key in self.keys():
            if key not in kwargs:
                raise DynamicRecArrayException(f"missing key {key} in arguments")
            if kwargs[key].shape[0] != kwargs[key0].shape[0]:
                raise DynamicRecArrayException(
                    f"array {key} size ({kwargs[key0].shape[0]}) != {key0} size ({kwargs[key0].shape[0]})"
                )
        add_len = kwargs[key0].shape[0]
        if add_len > 0:
            last_pos = len(self)
            data = np.resize(self.data, last_pos + add_len)
            for key in self.keys():
                data[key][last_pos:] = kwargs[key]
            self.data = data

    @property
    def data(self) -> np.array:
        """Return the numpy array. Access to the data array will also shring the capacity
        to the length of the array. The returned array is not a copy but a internal view.

        :return: np.aarry of type dtype
        """
        if self._capacity != self._length:
            self._array = self._array[:][: self._length]
            self._capacity = len(self._array)
        return self._array

    @data.setter
    def data(self, data):
        """Set the data of the record

        :param data: numpy array with the same dtype
        """
        self._length = len(data)
        self._capacity = len(data)
        self._array = data
